﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization.Charting;
using System.Globalization;

namespace Dcapture
{
    public partial class DcaptureControl : UserControl
    {
        public DcaptureControl()
        {
            
        }

        public void Init()
        {
            this.photonNeutronSeparation = new PhotonNeutronSeparation(MatrixCSV.EnergySizeMatrix, MatrixCSV.DiscrParamSizeMatrix);
            mainMatrix = new MatrixCSV(this.photonNeutronSeparation);
            shadowMatrix = new MatrixCSV(this.photonNeutronSeparation);
            

            InitializeComponent();

            graphControl.Initiate(mainMatrix, shadowMatrix, photonNeutronSeparation);

            this.udpReceiver = new UDPreceiver(mainMatrix);
            this.udpReceiver.UDPreceiverMessage += new UDPreceiver.UPDreceiverMessageHandler(udpReceiver_UDPreceiverMessage);

            graphControl.MouseMoveInGraph += new GraphControl.MouseMoveInGraphHandler(graphControl_MouseMoveInGraph);
            graphControl.SectionPreviewREDmove += new GraphControl.SectionPreviewMoveHandler(graphControl_SectionPreviewREDmove);
            graphControl.MouseClickInGraph += new GraphControl.MouseClickInGraphHandler(graphControl_MouseClickInGraph);

            udpReceiver.UDPreceiverError += new UDPreceiver.UDPreceiverErrorHandler(udpReceiver_UDPreceiverError);
        }


        void udpReceiver_UDPreceiverError(string message)
        {
            portNumber_numericUpDown.Enabled = true;
            UDPstart_button.Enabled = true;
            UDPstop_button.Enabled = false;

            label_UDPstate.ForeColor = SystemColors.ControlText;
            label_UDPstate.Text = "STOPPED";

            MessageBox.Show(message, "UDP receiver error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        void graphControl_MouseClickInGraph(int energy, int discrParam, MouseButtons mouseButton)
        {
            if (checkBoxSetPoints.Checked)
            {
                if (mouseButton == MouseButtons.Left)
                {
                    photonNeutronSeparation.SetTopPoint(energy, discrParam);
                }
                else if (mouseButton == MouseButtons.Right)
                {
                    photonNeutronSeparation.SetBottomPoint(energy, discrParam);
                }

                mainMatrix.ComputeSumOfParticles();
                this.ActualizeValues(true, false);
            }
        }

        /// <summary>
        /// UDP receiver changed the state
        /// </summary>
        /// <param name="message"></param>
        void udpReceiver_UDPreceiverMessage(UDPreceiver.UDPreceiverMessageEnum message)
        {
            if (message == UDPreceiver.UDPreceiverMessageEnum.newPacket)
            {
                ActualizeValues(false, true);
            }
            else if (message == UDPreceiver.UDPreceiverMessageEnum.noChange)
            {
                ActualizeValues(false, true);
            }
            else if (message == UDPreceiver.UDPreceiverMessageEnum.running)
            {
                portNumber_numericUpDown.Enabled = false;
                UDPstart_button.Enabled = false;
                UDPstop_button.Enabled = true;

                label_UDPstate.ForeColor = Color.ForestGreen;
                label_UDPstate.Text = "RUNNING";
            }
            else if (message == UDPreceiver.UDPreceiverMessageEnum.stopped)
            {
                portNumber_numericUpDown.Enabled = true;
                UDPstart_button.Enabled = true;
                UDPstop_button.Enabled = false;

                label_UDPstate.ForeColor = SystemColors.ControlText;
                label_UDPstate.Text = "STOPPED";
            }
            else
            {
                throw new Exception("new state");
            }
        }

        /// <summary>
        /// Show values while mose movement on graph
        /// </summary>
        /// <param name="energy"></param>
        /// <param name="discrParam"></param>
        void graphControl_MouseMoveInGraph(int energy, int discrParam)
        {
            energy_label.Text = energy.ToString();
            discrParam_label.Text = discrParam.ToString();
            try
            {
                mainMatrixImpulseCount_label.Text = mainMatrix[discrParam, energy].ToString();
                shadowMatrixImpulseCount_label.Text = shadowMatrix[discrParam, energy].ToString();
            }
            catch
            { 
                //indexes are out of the range
                energy_label.Text = "";
                discrParam_label.Text = "";
                mainMatrixImpulseCount_label.Text = "";
                shadowMatrixImpulseCount_label.Text = "";
            }
        }

        UDPreceiver udpReceiver = null;

        private void UDPstart_button_Click(object sender, EventArgs e)
        {
            try
            {
                udpReceiver.Start((int)portNumber_numericUpDown.Value);
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
            }

            portNumber_numericUpDown.Enabled = false;
            UDPstart_button.Enabled = false;
            UDPstop_button.Enabled = true;

            label_UDPstate.ForeColor = Color.Orange;
            label_UDPstate.Text = "STARTING";
        }

        MatrixCSV mainMatrix = null;
        MatrixCSV shadowMatrix = null;
        PhotonNeutronSeparation photonNeutronSeparation = null;

        private void OpenMatrix(MatrixCSV matrix)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.CheckFileExists = true;
            ofd.InitialDirectory = @"C:\Users\Martin\Desktop\SSD\output";
            ofd.Filter = "CSV files (*.csv)|*.csv|All files (*.*)|*.*";

            if (ofd.ShowDialog() != DialogResult.OK)
            {
                return;
            }

            try
            {
                matrix.AddFromCSV(ofd.FileName, true);
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
                return;
            }
            ActualizeValues(false, false);
        }

        private void buttonOpenCSV_colored_Click(object sender, EventArgs e)
        {
            OpenMatrix(mainMatrix);
        }

        private void openShadow_button_Click(object sender, EventArgs e)
        {
            OpenMatrix(shadowMatrix);
        }

        private void SaveMatrix(MatrixCSV matrix)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.OverwritePrompt = true;
            sfd.InitialDirectory = Application.StartupPath;
            sfd.Filter = "CSV files (*.csv)|*.csv";

            if (sfd.ShowDialog() != DialogResult.OK)
            {
                return;
            }

            try
            {
                matrix.SaveToCSV(sfd.FileName);
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
                return;
            }
            ActualizeValues(false, false);
        }

        private void saveMain_button_Click(object sender, EventArgs e)
        {
            SaveMatrix(mainMatrix);
        }

        uint prevCatchedPackets = 0;
        uint prevLostPackets = 0;
        uint prevSamplesOutOfMatrix = 0;
        uint prevCatchedSamples = 0;

        /// <summary>
        /// Actualize data on screen
        /// </summary>
        private void ActualizeValues(bool redrawOnlySectionPreview, bool actualizeSpeed)
        {
            //total samples in matrix
            totalSamplesMain_label.Text = mainMatrix.TotalSamples.ToString();
            totalSamplesShadow_label.Text = shadowMatrix.TotalSamples.ToString();

            //catched and lost samples
            cachedPackets_label.Text = udpReceiver.CatchedPackets.ToString();
            catchedSamples_label.Text = udpReceiver.CatchedSamples.ToString();
            lostPackets_label.Text = udpReceiver.LostPackets.ToString();
            labelSamplesOutOfMatrix.Text = udpReceiver.SamplesOutOfMatrix.ToString();

            //speed
            if (actualizeSpeed)
            {
                label_speedCatchedPackets.Text = (udpReceiver.CatchedPackets - prevCatchedPackets).ToString();
                label_speedCatchedSamples.Text = (udpReceiver.CatchedSamples - prevCatchedSamples).ToString();
                label_speedLostPackets.Text = (udpReceiver.LostPackets - prevLostPackets).ToString();
                label_speedSamplesOutOfMatrix.Text = (udpReceiver.SamplesOutOfMatrix - prevSamplesOutOfMatrix).ToString();

                prevCatchedPackets = udpReceiver.CatchedPackets;
                prevCatchedSamples = udpReceiver.CatchedSamples;
                prevLostPackets = udpReceiver.LostPackets;
                prevSamplesOutOfMatrix = udpReceiver.SamplesOutOfMatrix;
            }
            
            //start time and duration of UDP receiver
            startTime_label.Text = udpReceiver.StartTime.ToString("HH:mm:ss");
            TimeSpan duration = (DateTime.Now - udpReceiver.StartTime);
            duration_label.Text = duration.Hours.ToString("D2") + ":" + duration.Minutes.ToString("D2") + ":" + duration.Seconds.ToString("D2");

            graphControl.RedrawGraphRequest(redrawOnlySectionPreview);

            //photon / neutron recognition
            this.ShowSumOfParticles();
        }

        public void NeutronDetectionInit()
        {
            checkBoxNeutronDetection.Checked = true;

            for (int i = 0; i < historySumInC.Length; i++)
            {
                historySumInC[i] = 0;
                historySumInD[i] = 0;
            }

            labelRedAlarm.Visible = false;
            labelOrangeAlarm.Visible = false;
            labelStopBelt.Visible = false;
            neutronHistoryPointer = 0;
            
            this.neutronDetectionMessage.neutronAlarm = NeutronDetectionMessage.NeutronDetectionAlarm.OK;
            this.neutronDetectionMessage.stopBelt = false;
            
            mainMatrix.Clear();
            timerNeutronDetection.Start();
        }

        public void NeutronDetectionStop()
        {
            timerNeutronDetection.Stop();
            checkBoxNeutronDetection.Checked = false;
        }

        public event NeutronDetectionChangeStatusEventHandler NeutronDetectionChangeStatus;

        int neutronDetectionRunningLabel = 0;
        ulong[] historySumInC = new ulong[5];
        ulong[] historySumInD = new ulong[5];
        int neutronHistoryPointer = 0;
        NeutronDetectionMessage neutronDetectionMessage = new NeutronDetectionMessage();
        private void NeutronDetection(ulong actualSumInC, ulong actualSumInD)
        {       
            // '/' and '\' indicator
            neutronDetectionRunningLabel = (neutronDetectionRunningLabel + 1) % 2;
            labelNeutronDetectionRunning.Text = (neutronDetectionRunningLabel == 0) ? "/" : "\\";

            //write actual data
            historySumInC[neutronHistoryPointer] = actualSumInC;
            historySumInD[neutronHistoryPointer] = actualSumInD;
            
            //write history into textbox
            string text =  "      sum(C)  sum(D)      C:D\r\n";
            for (int i = 0; i < historySumInC.Length; i++)
            {
                int p = (neutronHistoryPointer - i + historySumInC.Length) % historySumInC.Length;

                float ratio = (historySumInD[p] / (float)(historySumInC[p] + historySumInD[p])) * 1000;
                text +=
                        "t=" + (-i).ToString().PadLeft(2, ' ') + "  " +
                        historySumInC[p].ToString().PadLeft(6, ' ') + "  " +
                        historySumInD[p].ToString().PadLeft(6, ' ') + "  " +
                        ((1000 - ratio).ToString("00.0", CultureInfo.InvariantCulture) + " : " + ratio.ToString("00.0", CultureInfo.InvariantCulture)).PadLeft(9, ' ') + "\r\n"; 
            }
            textBoxNeutronDetectionValues.Text = text;

            //compute and show alarm - it is not possible to downgrade the alarm - olny by reseting
            float actualRatio = (actualSumInD / (float)(actualSumInC + actualSumInD)) * 1000;
            if (actualSumInD > numericMinInD_red.Value && actualRatio > (float)numericMinRation_Red.Value) //red alarm
            {
                labelRedAlarm.Visible = true;
                labelOrangeAlarm.Visible = false;

                this.neutronDetectionMessage.neutronAlarm = NeutronDetectionMessage.NeutronDetectionAlarm.redAlarm;
            }
            else if ((this.neutronDetectionMessage.neutronAlarm != NeutronDetectionMessage.NeutronDetectionAlarm.redAlarm) && (actualSumInD > numericMinInD_orange.Value && actualRatio > (float)numericMinRation_Orange.Value)) //orange alarm
            {
                labelRedAlarm.Visible = false;
                labelOrangeAlarm.Visible = true;

                this.neutronDetectionMessage.neutronAlarm = NeutronDetectionMessage.NeutronDetectionAlarm.orangeAlarm;
            }

            //stop belt
            int pointerTminus1 = (neutronHistoryPointer - 1 + historySumInC.Length) % historySumInC.Length;
            if (this.neutronDetectionMessage.neutronAlarm == NeutronDetectionMessage.NeutronDetectionAlarm.orangeAlarm || this.neutronDetectionMessage.neutronAlarm == NeutronDetectionMessage.NeutronDetectionAlarm.redAlarm)
            {
                if (historySumInD[neutronHistoryPointer] < historySumInD[pointerTminus1])
                {
                    this.neutronDetectionMessage.stopBelt = true;
                    labelStopBelt.Visible = true;
                    NeutronDetectionStop();
                }
            }

            //clear all data -> prepare to new measuring interval
            mainMatrix.Clear();

            //add history pointer
            neutronHistoryPointer = (neutronHistoryPointer + 1) % historySumInC.Length;

            //make event
            if (NeutronDetectionChangeStatus != null)
            {
                NeutronDetectionChangeStatus(this.neutronDetectionMessage);
            }
        }

        private void ShowSumOfParticles()
        {
            ulong A = mainMatrix.ParticlesInA;
            ulong B = mainMatrix.ParticlesInB;
            ulong C = mainMatrix.ParticlesInC;
            ulong D = mainMatrix.ParticlesInD;
            ulong E = mainMatrix.ParticlesInE;
            ulong F = mainMatrix.ParticlesInF;

            textBoxSumParticlesBottomLeft.Text =
                "A = " + A.ToString() + "\r\n" +
                "C = " + C.ToString() + "\r\n" +
                "E = " + E.ToString();

            textBoxSumParticlesBottomRight.Text =
                "B = " + B.ToString() + "\r\n" +
                "D = " + D.ToString() + "\r\n" +
                "F = " + F.ToString();

            textBoxSumParticlesTop.Text =
                "        A:B = " + ((A / (float)(A + B)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + " : " +
                             ((B / (float)(A + B)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + "\r\n" +

                "        C:D = " + ((C / (float)(C + D)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + " : " +
                             ((D / (float)(C + D)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + "\r\n" +

                "        E:F = " + ((E / (float)(E + F)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + " : " +
                             ((F / (float)(E + F)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + "\r\n\r\n" +


                "    A+C:B+D = " + (((A + C) / (float)(A + B + C + D)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + " : " +
                                     (((B + D) / (float)(A + B + C + D)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + "\r\n" +

                "    C+E:D+F = " + (((C + E) / (float)(C + D + E + F)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + " : " +
                                     (((D + F) / (float)(C + D + E + F)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + "\r\n\r\n" +

                "A+C+E:B+D+F = " + (((A + C + E) / (float)(A + B + C + D + E + F)) * 100).ToString("00.0", CultureInfo.InvariantCulture) + " : " +
                                     (((B + D + F) / (float)(A + B + C + D + E + F)) * 100).ToString("00.0", CultureInfo.InvariantCulture);
        }

        private void UDPstop_button_Click(object sender, EventArgs e)
        {
            udpReceiver.Stop();

            label_UDPstate.ForeColor = Color.Orange;
            label_UDPstate.Text = "STOPPING";
        }

        private void clearMain_button_Click(object sender, EventArgs e)
        {
            mainMatrix.Clear();
            ActualizeValues(false, false);
        }

        private void clearShadow_button_Click(object sender, EventArgs e)
        {
            shadowMatrix.Clear();
            ActualizeValues(false, false);
        }

        private void saveShadow_button_Click(object sender, EventArgs e)
        {
            SaveMatrix(shadowMatrix);
        }

        private void AddMain_button_Click(object sender, EventArgs e)
        {
            shadowMatrix.AddMatrix(mainMatrix);
            ActualizeValues(false, false);
        }

        private void AddShadow_button_Click(object sender, EventArgs e)
        {
            mainMatrix.AddMatrix(shadowMatrix);
            ActualizeValues(false, false);
        }

        SpectrumForm spectrumFormR = null;
        SpectrumForm bothSpectraForm = null;

        /// <summary>
        /// Drawing position of the shadow matrix
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void shadowDoNotShow_radioButton_CheckedChanged(object sender, EventArgs e)
        {
            graphControl.SetShadowMatrixDrawingStyle(GraphControl.ShadowMatrixDrawingStyle.hide);
            this.ActualizeValues(false, false);
        }

        /// <summary>
        /// Drawing position of the shadow matrix
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void shadowBelow_radioButton_CheckedChanged(object sender, EventArgs e)
        {
            graphControl.SetShadowMatrixDrawingStyle(GraphControl.ShadowMatrixDrawingStyle.below);
            this.ActualizeValues(false, false);
        }

        /// <summary>
        /// Drawing position of the shadow matrix
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void shadowAbove_radioButton_CheckedChanged(object sender, EventArgs e)
        {
            graphControl.SetShadowMatrixDrawingStyle(GraphControl.ShadowMatrixDrawingStyle.above);
            this.ActualizeValues(false, false);
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            this.ActualizeValues(false, false);
        }

        /// <summary>
        /// Are you sure to close the window?
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (MessageBox.Show("Are you sure?", "Window closing", MessageBoxButtons.OKCancel, MessageBoxIcon.Question) == DialogResult.Cancel)
            {
                e.Cancel = true;
            }
        }

        private void sectionPreviewR_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            if (sectionPreviewR_checkBox.Checked)
            {
                mainMatrix.sectionPreviewR = 200;

                spectrumFormR = new SpectrumForm();
                spectrumFormR.FormClosed += new FormClosedEventHandler(spectrumFormR_FormClosed);
                spectrumFormR.Text = "RED Section preview";
                spectrumFormR.Show();
                ShowSectionPreview(mainMatrix.sectionPreviewR, spectrumFormR, Color.Red);
            }
            else if (spectrumFormR != null)
            {
                spectrumFormR.Close();
                spectrumFormR = null;
                mainMatrix.sectionPreviewR = -200;
            }

            this.ActualizeValues(true, false);
        }

        void spectrumFormR_FormClosed(object sender, FormClosedEventArgs e)
        {
            spectrumFormR = null;
            mainMatrix.sectionPreviewR = -200;
            sectionPreviewR_checkBox.Checked = false;
        }

        void graphControl_SectionPreviewREDmove()
        {
            if (spectrumFormR != null)
            {
                ShowSectionPreview(mainMatrix.sectionPreviewR, spectrumFormR, Color.Red);    
            }
        }

        /// <summary>
        /// Show 2D graph with section preview (only one energy, all discrim parameter)
        /// </summary>
        /// <param name="position"></param>
        /// <param name="spectrumForm"></param>
        /// <param name="name"></param>
        private void ShowSectionPreview(int position, SpectrumForm spectrumForm, Color color)
        {
            Series serie = spectrumFormR.chart.Series[0];
            serie.Points.Clear();
            serie.Color = color;

            for (int i = 0; i < mainMatrix.FurthermostSampleDiscrParam; i++)
            {
                serie.Points.Add(new DataPoint(i, mainMatrix[i, position]));
            }

        }

        private void showBothSpectra_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            if (showBothSpectra_checkBox.Checked)
            {
                bothSpectraForm = new SpectrumForm();
                bothSpectraForm.FormClosed += new FormClosedEventHandler(bothSpectraForm_FormClosed);
                bothSpectraForm.Show();

                Series serie = bothSpectraForm.chart.Series[0];
                serie.Name = "Photons and neutrons spectra together";

                int endDP = MatrixCSV.DiscrParamSizeMatrix;
                for (int j = 0; j < mainMatrix.FurthermostSampleEnergy; j++)
                {
                    ulong sum = 0;
                    for (int i = 0; i < endDP; i++)
                    {
                        sum += mainMatrix[i, j];
                    }
                    serie.Points.Add(new DataPoint(j, sum));
                }
            }
            else if (bothSpectraForm != null)
            {
                bothSpectraForm.Close();
                bothSpectraForm = null;
            }
        }

        void bothSpectraForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            bothSpectraForm = null;
            showBothSpectra_checkBox.Checked = false;
        }

        private void checkBoxNeutronDetection_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxNeutronDetection.Checked)
            {
                NeutronDetectionInit();
            }
            else
            {
                NeutronDetectionStop();
            }
        }

        private void timerNeutronDetection_Tick(object sender, EventArgs e)
        {
            
            NeutronDetection(mainMatrix.ParticlesInC, mainMatrix.ParticlesInD);


            //DEBUG
            //Random rnd = new Random();
            //NeutronDetection((ulong)rnd.Next(0, 1000), (ulong)rnd.Next(0, 10));
        }
    }


    public class NeutronDetectionMessage
    {
        public enum NeutronDetectionAlarm
        {
            OK,
            orangeAlarm,
            redAlarm
        }
        public NeutronDetectionAlarm neutronAlarm = NeutronDetectionAlarm.OK;

        public bool stopBelt = false;
    }

    public delegate void NeutronDetectionChangeStatusEventHandler(NeutronDetectionMessage message);
}
