﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace Dcapture
{
    public partial class GraphControl : UserControl
    {
        public GraphControl()
        {
            InitializeComponent();

            panel.MouseWheel += new MouseEventHandler(panel1_MouseWheel);
            graphName_textBox.MouseWheel += new MouseEventHandler(panel1_MouseWheel);

            bwGraphDrawer.DoWork += new DoWorkEventHandler(bwGraphDrawer_DoWork);
            bwGraphDrawer.RunWorkerCompleted += new RunWorkerCompletedEventHandler(bwGraphDrawer_RunWorkerCompleted);

            bwGraphDrawerWait.DoWork += new DoWorkEventHandler(bwGraphDrawerWait_DoWork);
            bwGraphDrawerWait.RunWorkerCompleted += new RunWorkerCompletedEventHandler(bwGraphDrawerWait_RunWorkerCompleted);
        }

        BackgroundWorker bwGraphDrawer = new BackgroundWorker();
        BackgroundWorker bwGraphDrawerWait = new BackgroundWorker();

        void panel1_MouseWheel(object sender, MouseEventArgs e)
        {
            int mult = e.Delta > 0 ? -1 : 1;
            
            if (Control.ModifierKeys == Keys.Shift)
            {
                //horizontal moving
                hScrollBarDiscrParam.Value = Math.Max(hScrollBarDiscrParam.Minimum, Math.Min(hScrollBarDiscrParam.Maximum - hScrollBarDiscrParam.LargeChange, hScrollBarDiscrParam.Value + (hScrollBarDiscrParam.LargeChange / 4 * mult)));
            }
            else if (Control.ModifierKeys == Keys.None)
            {
                //vertical moving
                vScrollBarEnergy.Value = Math.Max(vScrollBarEnergy.Minimum, Math.Min(vScrollBarEnergy.Maximum - vScrollBarEnergy.LargeChange, vScrollBarEnergy.Value + (vScrollBarEnergy.LargeChange / 4 * mult)));
            }
            else if (Control.ModifierKeys == Keys.Control)
            { 
                //verticalScroll zoom
                trackBarEnergyZoom.Value = Math.Max(trackBarEnergyZoom.Minimum, Math.Min(trackBarEnergyZoom.Maximum, trackBarEnergyZoom.Value + mult * -1));
            }
            else if (Control.ModifierKeys == (Keys.Shift|Keys.Control))
            { 
                //horizontal zoom
                trackBarDiscrParamZoom.Value = Math.Max(trackBarDiscrParamZoom.Minimum, Math.Min(trackBarDiscrParamZoom.Maximum, trackBarDiscrParamZoom.Value + mult * -1));
            }

            RedrawGraphRequest(false);
        }


        int bitmapSizeInBits = 9;
        public int BitmapSize
        {
            get { return (int)Math.Pow(2, bitmapSizeInBits); }
        }

        GraphDrawer graphicDrawer = null;
        MatrixCSV mainMatrix = null;
        MatrixCSV shadowMatrix = null;
        PhotonNeutronSeparation photonNeutronSeparation = null;

        public void Initiate(MatrixCSV mainMatrix, MatrixCSV shadowMatrix, PhotonNeutronSeparation photonNeutronSeparation)
        {
            this.mainMatrix = mainMatrix;
            this.shadowMatrix = shadowMatrix;
            this.photonNeutronSeparation = photonNeutronSeparation;

            this.graphicDrawer = new GraphDrawer(9, mainMatrix, shadowMatrix, photonNeutronSeparation);    //bitmap dimension is 512x512
        }
        
        private void EnergyCoefs(int Y, out int energy, out int energyMin, out int energyMax)
        {
            energy = vScrollBarEnergy.Value + Y * this.energy_divForActualZoom / this.energy_zoomBitmap;
            energyMin = vScrollBarEnergy.Value + (Y - 5) * this.energy_divForActualZoom / this.energy_zoomBitmap;
            energyMax = vScrollBarEnergy.Value + (Y + 5) * this.energy_divForActualZoom / this.energy_zoomBitmap;
        }

        private int DiscrCoef(int X)
        {
            return hScrollBarDiscrParam.Value + X * this.discrParam_divForActualZoom / this.discrParam_zoomBitmap;
        }

        /// <summary>
        /// start moving the section preview
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void pictureBoxGraph_MouseDown(object sender, MouseEventArgs e)
        {
            int energy;
            int energyMin;
            int energyMax;
            EnergyCoefs(e.Y, out energy, out energyMin, out energyMax);

            if (energyMin < mainMatrix.sectionPreviewR && energyMax > mainMatrix.sectionPreviewR && e.Button == MouseButtons.Left)
            {
                movingR = true;
                pictureBoxGraph.Cursor = Cursors.SizeNS;
            }

            movedR = false;
        }

        /// <summary>
        /// stop moving the section preview
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void pictureBoxGraph_MouseUp(object sender, MouseEventArgs e)
        {
            movingR = false;

            int energy;
            int energyMin;
            int energyMax;
            EnergyCoefs(e.Y, out energy, out energyMin, out energyMax);

            int discrParam = DiscrCoef(e.X);

            if (MouseClickInGraph != null && !movedR)
            {
                MouseClickInGraph(energy, discrParam, e.Button);
            }
        }

        bool movingR = false;
        bool movedR = false;
        public delegate void SectionPreviewMoveHandler();
        public event SectionPreviewMoveHandler SectionPreviewREDmove;

        private int previousSectionPreviewR = 0;

        private void pictureBoxGraph_MouseMove(object sender, MouseEventArgs e)
        {
            int energy;
            int energyMin;
            int energyMax;
            EnergyCoefs(e.Y, out energy, out energyMin, out energyMax);

            int discrParam = DiscrCoef(e.X);

            //position of section preview
            if (movingR)
            {
                mainMatrix.sectionPreviewR = Math.Min(MatrixCSV.EnergySizeMatrix - 1, Math.Max(0, energy));
                RedrawGraphRequest(true);

                //make event
                if (SectionPreviewREDmove != null && this.previousSectionPreviewR != mainMatrix.sectionPreviewR)
                {
                    SectionPreviewREDmove();
                    movedR = true;
                }

                //stop moving section preview
                if (e.Button != MouseButtons.Left)
                {
                    movingR = false;
                }

                this.previousSectionPreviewR = mainMatrix.sectionPreviewR;
            }
            else
            {
                if (energyMin < mainMatrix.sectionPreviewR && energyMax > mainMatrix.sectionPreviewR)
                {
                    pictureBoxGraph.Cursor = Cursors.SizeNS;
                }
                else
                {
                    pictureBoxGraph.Cursor = Cursors.Default;
                }
            }

            
            //make event while mouse movement in graph
            if (MouseMoveInGraph != null)
            {
                MouseMoveInGraph(energy, discrParam);
            }
        }

        public delegate void MouseMoveInGraphHandler(int energy, int discrParam);
        public event MouseMoveInGraphHandler MouseMoveInGraph;

        public delegate void MouseClickInGraphHandler(int energy, int discrParam, MouseButtons mouseButton);
        public event MouseClickInGraphHandler MouseClickInGraph;

        private void pictureBoxGraph_MouseClick(object sender, MouseEventArgs e)
        {
            //enable scrooling by mouse wheel
            panel.Focus();
        }

        /// <summary>
        /// Scrool the graph (x and y)
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ScrollBars_Scroll(object sender, ScrollEventArgs e)
        {
            RedrawGraphRequest(false);
        }

        /// <summary>
        /// Zoom the graph (x and y)
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void trackBarsZoom_Scroll(object sender, EventArgs e)
        {
            RedrawGraphRequest(false);
        }

        //output parameters
        //energyGraph_startIndex = energy_startIndex + positionEnergy;
        //energyGraph_endIndex = energy_endIndex + positionEnergy - 1;

        public enum ShadowMatrixDrawingStyle
        {
            hide,
            below,
            above,
            watermark
        }

        private ShadowMatrixDrawingStyle shadowMatrixDrawingStyle = ShadowMatrixDrawingStyle.hide;
        public void SetShadowMatrixDrawingStyle(ShadowMatrixDrawingStyle shadowMatrixDrawingStyle)
        {
            this.shadowMatrixDrawingStyle = shadowMatrixDrawingStyle;
        }

        int energy_divForActualZoom;
        int energy_zoomBitmap;
        int energy_endIndex;
        int positionEnergy;

        int discrParam_divForActualZoom;
        int discrParam_zoomBitmap;
        int discrParam_endIndex;
        int positionDiscrParam;
        
        /// <summary>
        /// call RedrawGraph() in small time intervals
        /// </summary>
        public void RedrawGraphRequest(bool redrawOnlySectionPreview)
        {
            if (!drawingGraphBusy)
            {
                this.RedrawGraphRequest_redrawWholeGraph = false;
                this.RedrawGraphRequest_storage = false;
                RedrawGraph_PartA(redrawOnlySectionPreview);
            }
            else
            {
                this.RedrawGraphRequest_storage = true;
                if (!redrawOnlySectionPreview)
                {
                    this.RedrawGraphRequest_redrawWholeGraph = true;
                }
            }
        }

        private bool RedrawGraphRequest_storage = false;
        private bool RedrawGraphRequest_redrawWholeGraph = false;
        bool drawingGraphBusy = false;
        
        
        /// <summary>
        /// Sets the scroll bars and draw the graph
        /// </summary>
        private void RedrawGraph_PartA(bool redrawOnlySectionPreview)
        {
            drawingGraphBusy = true;
            labelDrawing.Visible = true;
            
            //compute params to make one matrix - energy axe
            ComputeCoeficients(MatrixCSV.EnergySizeMatrix, BitmapSize, trackBarEnergyZoom.Value, out energy_divForActualZoom, out energy_zoomBitmap, out energy_endIndex);

            //compute params to make one matrix - discrimParam axe
            ComputeCoeficients(MatrixCSV.DiscrParamSizeMatrix, BitmapSize, trackBarDiscrParamZoom.Value, out discrParam_divForActualZoom, out discrParam_zoomBitmap, out discrParam_endIndex);




            //draw graph
            this.positionEnergy = vScrollBarEnergy.Value;
            this.positionDiscrParam = hScrollBarDiscrParam.Value;

            GraphDrawer.GrawGraphArguments args = new GraphDrawer.GrawGraphArguments();
            args.positionDiscrParam = positionDiscrParam;
            args.discrParam_endIndex = discrParam_endIndex;
            args.discrParam_divForActualZoom = discrParam_divForActualZoom;
            args.discrParam_zoomBitmap = discrParam_zoomBitmap;
            args.positionEnergy = positionEnergy;
            args.energy_endIndex = energy_endIndex;
            args.energy_divForActualZoom = energy_divForActualZoom;
            args.energy_zoomBitmap = energy_zoomBitmap;
            args.shadowMatrixDrawingStyle = shadowMatrixDrawingStyle;
            args.colorSaturation = colorSaturation_trackBar.Value;
            args.redrawOnlySectionPreview = redrawOnlySectionPreview;

            bwGraphDrawer.RunWorkerAsync(args);
        }

        void bwGraphDrawer_DoWork(object sender, DoWorkEventArgs e)
        {
            e.Result = graphicDrawer.DrawGraph((GraphDrawer.GrawGraphArguments)e.Argument);
        }


        void bwGraphDrawer_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            RedrawGraph_PartB((Image)e.Result);
        }


        private void RedrawGraph_PartB(Image graph)
        {
            pictureBoxGraph.Image = graph;

            //set scrool bar - energy
            int maxFurhermosSampleEnergy = Math.Max(mainMatrix.FurthermostSampleEnergy, shadowMatrix.FurthermostSampleEnergy);
            int maxPositionEnergy = Math.Max(0, Math.Min(MatrixCSV.EnergySizeMatrix, (maxFurhermosSampleEnergy * 11 / 10)) - energy_endIndex);
            int smallChangeEnergy = energy_divForActualZoom;
            int largeChangeEnergy = energy_endIndex;

            if (maxPositionEnergy == 0)
            {
                vScrollBarEnergy.Value = 0;
                vScrollBarEnergy.Enabled = false;
            }
            else
            {
                vScrollBarEnergy.Enabled = true;
                vScrollBarEnergy.Maximum = maxPositionEnergy + largeChangeEnergy;
                vScrollBarEnergy.LargeChange = largeChangeEnergy;
                vScrollBarEnergy.SmallChange = smallChangeEnergy;
            }

            //set scroll bar - discrParam
            int maxFurthermostSampleDiscrParam = Math.Max(mainMatrix.FurthermostSampleDiscrParam, shadowMatrix.FurthermostSampleDiscrParam);
            int maxPositionDiscrParam = Math.Max(0, Math.Min(MatrixCSV.DiscrParamSizeMatrix, (maxFurthermostSampleDiscrParam * 11 / 10)) - discrParam_endIndex);
            int smallChangeDiscrParam = discrParam_divForActualZoom;
            int largeChangeDiscrParam = discrParam_endIndex;

            if (maxPositionDiscrParam == 0)
            {
                hScrollBarDiscrParam.Value = 0;
                hScrollBarDiscrParam.Enabled = false;
            }
            else
            {
                hScrollBarDiscrParam.Enabled = true;
                hScrollBarDiscrParam.Maximum = maxPositionDiscrParam + largeChangeDiscrParam;
                hScrollBarDiscrParam.LargeChange = largeChangeDiscrParam;
                hScrollBarDiscrParam.SmallChange = smallChangeDiscrParam;
            }

            //set energy axe
            int stepEnergy = energy_endIndex / 4;
            labelEnergy0.Text = positionEnergy.ToString();
            labelEnergy1.Text = (positionEnergy + 1 * stepEnergy).ToString();
            labelEnergy2.Text = (positionEnergy + 2 * stepEnergy).ToString();
            labelEnergy3.Text = (positionEnergy + 3 * stepEnergy).ToString();
            labelEnergy4.Text = (positionEnergy + 4 * stepEnergy).ToString();

            //set discrParam axe
            int stepDiscrParam = discrParam_endIndex / 4;
            labelDiscrParam0.Text = (positionDiscrParam).ToString();
            labelDiscrParam1.Text = (positionDiscrParam + 1 * stepDiscrParam).ToString();
            labelDiscrParam2.Text = (positionDiscrParam + 2 * stepDiscrParam).ToString();
            labelDiscrParam3.Text = (positionDiscrParam + 3 * stepDiscrParam).ToString();
            labelDiscrParam4.Text = (positionDiscrParam + 4 * stepDiscrParam).ToString();

            //set color axe
            int colorStep = (int)(graphicDrawer.MaxNumSamplesInGraph / 4);
            labelColor0.Text = (1).ToString();
            labelColor1.Text = (1 * colorStep).ToString();
            labelColor2.Text = (2 * colorStep).ToString();
            labelColor3.Text = (3 * colorStep).ToString();
            labelColor4.Text = (4 * colorStep).ToString();

            //mouse wheel will not be focused on the trackBar
            //panel.Focus();

            labelDrawing.Visible = false;
            bwGraphDrawerWait.RunWorkerAsync();
        }

        void bwGraphDrawerWait_DoWork(object sender, DoWorkEventArgs e)
        {
            System.Threading.Thread.Sleep(30);
        }

        void bwGraphDrawerWait_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            drawingGraphBusy = false;
            
            //check if the reques was during previous 
            if (this.RedrawGraphRequest_storage)
            {
                RedrawGraph_PartA(!this.RedrawGraphRequest_redrawWholeGraph);
                this.RedrawGraphRequest_redrawWholeGraph = false;
                this.RedrawGraphRequest_storage = false;
            }
        }

        private void ComputeCoeficients(int mainMatrixDimension, int bitmapDimension, int zoom, out int divActualZoom, out int zoomBitmap, out int endIndex)
        {
            int divForZoom1 = mainMatrixDimension / bitmapDimension;
            int zoomReal = (int)Math.Pow(2, zoom);
            divActualZoom = divForZoom1 / zoomReal;
            zoomBitmap = 1;
            if (divActualZoom == 0)
            {
                divActualZoom = 1;
                zoomBitmap = zoomReal / divForZoom1;
            }
            endIndex = bitmapDimension * divActualZoom / zoomBitmap;
        }

        private void colorSaturation_trackBar_Scroll(object sender, EventArgs e)
        {
            RedrawGraphRequest(false);
        }

        private void save_button_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.OverwritePrompt = true;
            sfd.FileName = graphName_textBox.Text;
            sfd.InitialDirectory = Application.StartupPath;
            sfd.Filter = "PNG files (*.png)|*.png";

            if (sfd.ShowDialog() != DialogResult.OK)
            {
                return;
            }

            try
            {
                Bitmap bmpScreenshot = new Bitmap(panel.Size.Width, panel.Size.Height);
                Graphics gfxScreenshot = Graphics.FromImage(bmpScreenshot);
                Point upperLefCorner = panel.PointToScreen(new Point(0, 0));
                gfxScreenshot.CopyFromScreen(upperLefCorner.X-1, upperLefCorner.Y-1, 0, 0, panel.Size, CopyPixelOperation.SourceCopy);
                bmpScreenshot.Save(sfd.FileName, System.Drawing.Imaging.ImageFormat.Png);
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
                return;
            }
        }

        private void GraphControl_Load(object sender, EventArgs e)
        {
            graphName_textBox.Text = "Dcapture    " + DateTime.Now.ToLongDateString();
        }
    }
}
