﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Dcapture
{
    public class PhotonNeutronSeparation
    {
        /// <summary>
        /// points that define the border between photons and neutrons
        /// y axis points are 0, 1, 2, 3, ...
        /// X axis points are stored in this field
        /// </summary>
        int[] separationLine = null;

        public int[] SeparationLine
        {
            get
            {
                return separationLine;
            }
        }

        private int topDiscrParam = -1;
        private int topEnergy = -1;
        private int bottomDiscrParam = -1;
        private int bottomEnergy = -1;

        public int EnergyOfTopLine
        {
            get
            {
                return topEnergy;
            }
        }

        public int EnergyOfBottomLine
        {
            get
            {
                return bottomEnergy;
            }
        }

        private int maxDiscrParam = 0;
        
        public PhotonNeutronSeparation(int matrixSizeEnergy, int matrixSizeDiscrParam)
        { 
            this.separationLine = new int[matrixSizeEnergy];
            for (int i = 0; i < separationLine.Length; i++)
            {
                separationLine[i] = 0;
            }

            this.maxDiscrParam = matrixSizeDiscrParam - 1;

            ComputeSeparationLine();
        }

        public void SetTopPoint(int energy, int discrParam)
        {
            this.topEnergy = energy;
            this.topDiscrParam = discrParam;

            this.ComputeSeparationLine();
        }

        public void SetBottomPoint(int energy, int discrParam)
        {
            this.bottomEnergy = energy;
            this.bottomDiscrParam = discrParam;

            this.ComputeSeparationLine();
        }

        private void ComputeSeparationLine()
        { 
            //compute as a general equation of a line (ax + by + c = 0)
            
            //direction vector of a line
            int u1 = bottomDiscrParam - topDiscrParam;
            int u2 = bottomEnergy - topEnergy;

            //normal vector of line (n1, n2) and parameter a and b
            int a = u2;     //n1
            int b = -u1;    //n2

            //parameter c
            int c = -(a*topDiscrParam + b*topEnergy);

            //compute points
            for (int i = 0; i < separationLine.Length; i++)
            {
                long value = -1;
                if (a != 0)
                {
                    value = (-b * i - c) / a;
                }

                value = Math.Min(Math.Max(-1, value), maxDiscrParam+1);
                separationLine[i] = (int)value;
            }
        }
    }
}
